/*  Class Matrix
*
*   Defines a matrix and includes the methods
*   needed for standard matrix manipulations, e.g. multiplation,
*   and related procedures, e.g. solution of linear
*   simultaneous equations
*
*   See class ComplexMatrix for complex matrix arithmetic
*
*   WRITTEN BY: Dr Michael Thomas Flanagan
*
*   DATE:	 June 2002
*   UPDATE:  21 April 2004, 19 January 2005, 1 May 2005
*
*   DOCUMENTATION:
*   See Michael Thomas Flanagan's Java library on-line web page:
*   Matrix.html
*
*   Copyright (c) April 2004  Michael Thomas Flanagan
*
*   PERMISSION TO COPY:
*   Permission to use, copy and modify this software and its documentation for
*   NON-COMMERCIAL purposes is granted, without fee, provided that an acknowledgement
*   to the author, Michael Thomas Flanagan at www.ee.ucl.ac.uk/~mflanaga, appears in all copies.
*
*   Dr Michael Thomas Flanagan makes no representations about the suitability
*   or fitness of the software for any or for a particular purpose.
*   Michael Thomas Flanagan shall not be liable for any damages suffered
*   as a result of using, modifying or distributing this software or its derivatives.
*
***************************************************************************************/

/****************************************************************************
*
*		Modificaes no nome da classe introduzida por Jos Iguelmar Miranda,
*		bem como alguns mdulos de manipulao de matrizes, principalmente no que
*		se refere ao clculo de autovalores e autovetores.
*
*       Parte integrante do software 'Sofia Imagens - SOFtware Inteligente de Anlise 
*                                                     de Imagens'.
*
*****************************************************************************/

package pID;

// Pacotes genricos
import java.text.DecimalFormat;
import java.util.Arrays;

public class Matriz {

  // DATA VARIABLES

	private int nrow = 0;               	// number of rows
	private int ncol = 0;               	// number of columns
	private double matriz[][] = null; 	    // 2-D  Matriz
	private int index[] = null;            	// row permutation index
	private double dswap = 1.0D;           	// row swap index
	private boolean matrizCheck = true;     // check on matrix status
	                                	    // true - no problems encountered
	                                	    // false - attempted a LU decomposition on a singular matrix

	// Atributos para clculo de autovalores - Miranda
	// O array 'eigen' armazena resultados de eigen:
	// (1) eigenvalues na primeira linha e 
	// (2) eigenvectors da segunda linha em diante.
	//
	private double[][] eigen = null;

	private static final double TINY = 1.0e-30;

	// CONSTRUCTORS
	// Construct a nrow x ncol matrix of complex variables all equal to zero
	//
  public Matriz(int nrow, int ncol){
   	this.nrow = nrow;
		this.ncol = ncol;
		this.matriz = new double[nrow][ncol];
		this.index = new int[nrow];
		for (int i = 0; i < nrow; i++)
			this.index[i] = i;
  }

	// Construct a nrow x ncol matrix of complex variables all equal to 
	// the complex number const
	//
  public Matriz(int nrow, int ncol, double constant){
   	this.nrow = nrow;
		this.ncol = ncol;
		this.matriz = new double[nrow][ncol];
		for (int i = 0; i < nrow; i++){
			for (int j = 0; j < nrow; j++)
				this.matriz[i][j] = constant;
		}
		this.index = new int[nrow];
		for (int i = 0; i < nrow; i++)
			this.index[i] = i;
 	}

	// Construct matrix with a reference to an existing nrow x ncol 2-D 
	// array of complex variables
	//
  public Matriz(double[][] twoD){
		this.nrow = twoD.length;
		this.ncol = twoD[0].length;
		for (int i = 0; i < nrow; i++){
		    if (twoD[i].length != ncol)
		    	throw new IllegalArgumentException("All rows must have the same length");
		}
		this.matriz = twoD;
		this.eigen =  new double[nrow][ncol];		// Miranda
		this.index = new int[nrow];
		for (int i = 0; i < nrow; i++)
			this.index[i] = i;
	}

	// Construct matrix with a reference to the 2D matrix and 
	// permutation index of an existing ComplexMatriz bb.
	//
  public Matriz(Matriz bb){
		this.nrow = bb.nrow;
		this.ncol = bb.ncol;
		this.matriz = bb.matriz;
		this.index = bb.index;
		this.dswap = bb.dswap;
 	}

    
  // METHODS

  // SET VALUES
  // Set the matrix with a copy of an existing nrow x ncol 2-D matrix of 
	// complex variables
	//
  public void setTwoDarray(double[][] aarray){
   	String msgA = "row length of this ComplexMatriz differs from that of the 2D array argument";
		String msgB = "column length of this ComplexMatriz differs from that of the 2D array argument";
		String msgC = "All rows must have the same length";
		if (this.nrow != aarray.length)
			throw new IllegalArgumentException(msgA);
		if (this.ncol != aarray[0].length)
			throw new IllegalArgumentException(msgB);
		for (int i = 0; i < nrow; i++){
			if (aarray[i].length != ncol)
				throw new IllegalArgumentException(msgC);
			for (int j = 0; j < ncol; j++){
				this.matriz[i][j] = aarray[i][j];
			}
		}
	}

  // Set an individual array element
  // i = row index
  // j = column index
  // aa = value of the element
	//
  public void setElement(int i, int j, double aa){
   	this.matriz[i][j] = aa;
  }

  // Set a sub-matrix starting with column index i, row index j
  // and ending with column index k, row index l
	//
  public void setSubMatriz(int i, int j, int k, int l, double[][] smat){
   	if (i > k)
			throw new IllegalArgumentException("row indices inverted");
    if (j > l)
    	throw new IllegalArgumentException("column indices inverted");
    int n = k-i+1, m = j-l+1;
    for (int p = 0; p < n; p++){
    	for (int q = 0; q < m; p++){
    		this.matriz[i+p][j+q] = smat[i][j];
    	}
    }
  }

  // Set a sub-matrix
  // row = array of row indices
  // col = array of column indices
	//
  public void setSubMatriz(int[] row, int[] col, double[][] smat){
   	int n = row.length;
   	int m = col.length;
   	for (int p = 0; p < n; p++){
   		for (int q = 0; q < m; p++){
   			this.matriz[row[p]][col[q]] = smat[p][q];
   		}
   	}
  }

  // Get the value of matrixCheck
	//
  public boolean getMatrizCheck(){
   	return this.matrizCheck;
  }

  // SPECIAL MATRICES
  // Construct an identity matrix
	//
  public static Matriz identityMatriz(int nrow){
   	Matriz u = new Matriz(nrow, nrow);
   	for (int i = 0; i < nrow; i++){
   		u.matriz[i][i] = 1.0;
   	}
   	return u;
  }

  // Construct a complex scalar matrix
	//
  public static Matriz scalarMatriz(int nrow, double diagconst){
   	Matriz u = new Matriz(nrow, nrow);
   	double[][] uarray = u.getArrayReference();
   	for (int i = 0; i < nrow; i++){
   		for (int j = i; j < nrow; j++){
   			if (i == j){
   				uarray[i][j] =  diagconst;
   			}
   		}
   	}
   	return u;
  }

  // Construct a diagonal matrix
	//
  public static Matriz diagonalMatriz(int nrow, double[] diag){
   	String msgA = "matriz dimension differs from diagonal array length";
   	if (diag.length != nrow)
   		throw new IllegalArgumentException(msgA);
   	Matriz u = new Matriz(nrow, nrow);
   	double[][] uarray = u.getArrayReference();
   	for (int i = 0; i < nrow; i++){
   		uarray[i][i] = diag[i];
   	}
   	return u;
  }

  // GET VALUES
  // Return the number of rows
	//
  public int getNrow(){
   	return this.nrow;
  }

  // Return the number of columns
	//
  public int getNcol(){
   	return this.ncol;
  }

  // Return a reference to the internal 2-D array
	//
  public double[][] getArrayReference(){
   	return this.matriz;
  }

  // Return a reference to the internal 2-D array
  // included for backward compatibility with incorrect earlier documentation
	//
  public double[][] getArrayPointer(){
   	return this.matriz;
  }

  // Return a copy of the internal 2-D array
	//
  public double[][] getArrayCopy(){
  	double[][] c = new double[this.nrow][this.ncol];
		for (int i = 0; i < nrow; i++){
			for (int j = 0; j < ncol; j++){
				c[i][j] = this.matriz[i][j];
			}
		}
		return c;
  }

  // Return  a single element of the internal 2-D array
	//
  public double getElement(int i, int j){
   	return this.matriz[i][j];
  }

  // Return a single element of the internal 2-D array
  // included for backward compatibility with incorrect earlier documentation
	//
  public double getElementCopy(int i, int j){
  	return this.matriz[i][j];
  }

  // Return a single element of the internal 2-D array
  // included for backward compatibility with incorrect earlier documentation
	//
  public double getElementPointer(int i, int j){
   	return this.matriz[i][j];
  }

  // Return a sub-matrix starting with column index i, row index j
  // and ending with column index k, row index l
	//
  public Matriz getSubMatriz(int i, int j, int k, int l){
   	if (i > k)
  		throw new IllegalArgumentException("row indices inverted");
  	if (j > l)
   		throw new IllegalArgumentException("column indices inverted");
   	int n = k-i+1, m = j-l+1;
   	Matriz smat = new Matriz(n, m);
   	double[][] sarray = this.getArrayReference();
   	for (int p = 0; p < n; p++){
   		for (int q = 0; q < m; p++){
   			sarray[p][q] =  this.matriz[i+p][j+q];
   		}
   	}
   	return smat;
  }

  // Return a sub-matrix
  // row = array of row indices
  // col = array of column indices
	//
  public Matriz getSubMatriz(int[] row, int[] col){
   	int n = row.length;
   	int m = col.length;
   	Matriz smat = new Matriz(n, m);
  	double[][] sarray = this.getArrayReference();
   	for (int i = 0; i < n; i++){
   		for (int j = 0; j < m; j++){
   			sarray[i][j] =  this.matriz[row[i]][col[j]];
   		}
   	}
   	return smat;
  }

  // Return a reference to the permutation index array
	//
  public int[]  getIndexReference(){
   	return this.index;
  }

  // Return a reference to the permutation index array
  // included for backward compatibility with incorrect earlier documentation
	//
  public int[]  getIndexPointer(){
   	return this.index;
  }

  // Return a copy of the permutation index array
	//
  public int[]  getIndexCopy(){
   	int[] indcopy = new int[this.nrow];
   	for (int i = 0; i < this.nrow; i++){
   		indcopy[i] = this.index[i];
   	}
   	return indcopy;
  }

  // Return the row swap index
	//
  public double getSwap(){
   	return this.dswap;
  }

  // COPY
  // Copy a Matriz [static method]
	//
  public static Matriz copy(Matriz a){
   	if (a == null){
   		return null;
   	} else {
   		int nr = a.getNrow();
   		int nc = a.getNcol();
   		double[][] aarray = a.getArrayReference();
   		Matriz b = new Matriz(nr,nc);
   		b.nrow = nr;
   		b.ncol = nc;
   		double[][] barray = b.getArrayReference();
   		for (int i = 0; i < nr; i++){
   			for (int j = 0; j < nc; j++){
   				barray[i][j] = aarray[i][j];
   			}
   		}
   		for (int i = 0; i < nr; i++)
   			b.index[i] = a.index[i];
   		return b;
   	}
  }

  // Copy a Matriz [instance method]
	//
  public Matriz copy(){
   	if (this == null){
   		return null;
   	} else {
   		int nr = this.nrow;
   		int nc = this.ncol;
   		Matriz b = new Matriz(nr,nc);
   		double[][] barray = b.getArrayReference();
   		b.nrow = nr;
  		b.ncol = nc;
   		for (int i = 0; i < nr; i++){
   			for (int j = 0; j < nc; j++){
   				barray[i][j] = this.matriz[i][j];
   			}
   		}
   		for (int i = 0; i < nr; i++)
    			b.index[i] = this.index[i];
   		return b;
   	}
  }

  // Clone a Matriz
	//
  public Object clone(){
   	if (this == null){
   		return null;
   	} else {
   		int nr = this.nrow;
   		int nc = this.ncol;
   		Matriz b = new Matriz(nr,nc);
   		double[][] barray = b.getArrayReference();
   		b.nrow = nr;
   		b.ncol = nc;
   		for (int i = 0; i < nr; i++){
   			for (int j = 0; j < nc; j++){
   				barray[i][j] = this.matriz[i][j];
   			}
   		}
   		for (int i = 0; i < nr; i++)
   			b.index[i] = this.index[i];
   		return (Object) b;
   	}
  }

  // ADDITION
  // Add this matrix to matrix B.  This matrix remains unaltered 
	// [instance method]
	//
  public Matriz plus(Matriz bmat){
   	if ((this.nrow != bmat.nrow)||(this.ncol != bmat.ncol)){
   		throw new IllegalArgumentException("Array dimensions do not agree");
   	}
   	int nr = bmat.nrow;
   	int nc = bmat.ncol;
   	Matriz cmat = new Matriz(nr,nc);
   	double[][] carray = cmat.getArrayReference();
   	for (int i = 0; i < nr; i++){
   		for (int j = 0; j < nc; j++){
   			carray[i][j] = this.matriz[i][j] + bmat.matriz[i][j];
   		}
   	}
   	return cmat;
  }

  // Add matrices A and B [static method]
	//
  public static Matriz plus(Matriz amat, Matriz bmat){
   	if ((amat.nrow != bmat.nrow)||(amat.ncol != bmat.ncol)){
   		throw new IllegalArgumentException("Array dimensions do not agree");
   	}
   	int nr = amat.nrow;
   	int nc = amat.ncol;
   	Matriz cmat = new Matriz(nr,nc);
   	double[][] carray = cmat.getArrayReference();
   	for (int i = 0; i < nr; i++){
   		for (int j = 0; j < nc; j++){
   			carray[i][j] = amat.matriz[i][j] + bmat.matriz[i][j];
   		}
   	}
   	return cmat;
  }

  // Add matrix B to this matrix [equivalence of +=]
	//
  public void plusEquals(Matriz bmat){
   	if ((this.nrow != bmat.nrow)||(this.ncol != bmat.ncol)){
   		throw new IllegalArgumentException("Array dimensions do not agree");
   	}
   	int nr = bmat.nrow;
   	int nc = bmat.ncol;
    	for (int i = 0; i < nr; i++){
   		for (int j = 0; j < nc; j++){
   			this.matriz[i][j] += bmat.matriz[i][j];
   		}
   	}
	}

  // SUBTRACTION
  // Subtract matrix B from this matrix. This matrix remains unaltered 
	// [instance method]
	//
  public Matriz minus(Matriz bmat){
   	if ((this.nrow != bmat.nrow)||(this.ncol != bmat.ncol)){
   		throw new IllegalArgumentException("Array dimensions do not agree");
   	}
   	int nr = this.nrow;
   	int nc = this.ncol;
   	Matriz cmat = new Matriz(nr,nc);
   	double[][] carray = cmat.getArrayReference();
   	for (int i = 0; i < nr; i++){
   		for (int j = 0; j < nc; j++){
   			carray[i][j] = this.matriz[i][j] - bmat.matriz[i][j];
   		}
   	}
   	return cmat;
  }

  // Subtract matrix B from matrix A [static method]
	//
  public static Matriz minus(Matriz amat, Matriz bmat){
   	if ((amat.nrow != bmat.nrow)||(amat.ncol != bmat.ncol)){
   		throw new IllegalArgumentException("Array dimensions do not agree");
   	}
   	int nr = amat.nrow;
   	int nc = amat.ncol;
   	Matriz cmat = new Matriz(nr,nc);
   	double[][] carray = cmat.getArrayReference();
   	for (int i = 0; i < nr; i++){
   		for (int j = 0; j < nc; j++){
   			carray[i][j] = amat.matriz[i][j] - bmat.matriz[i][j];
   		}
   	}
   	return cmat;
  }

  // Subtract matrix B from this matrix [equivlance of -=]
	//
  public void minusEquals(Matriz bmat){
   	if ((this.nrow != bmat.nrow)||(this.ncol != bmat.ncol)){
   		throw new IllegalArgumentException("Array dimensions do not agree");
   	}
   	int nr = bmat.nrow;
   	int nc = bmat.ncol;
    	for (int i = 0; i < nr; i++){
   		for (int j = 0; j < nc; j++){
   			this.matriz[i][j] -= bmat.matriz[i][j];
   		}
   	}
  }

  // MULTIPLICATION
  // Multiply this  matrix by a matrix.   [instance method]
  // This matrix remains unaltered.
	//
  public Matriz times(Matriz bmat){
  	if (this.ncol != bmat.nrow)
   		throw new IllegalArgumentException("Nonconformable matrices");
   	Matriz cmat = new Matriz(this.nrow, bmat.ncol);
   	double[][] carray = cmat.getArrayReference();
   	double sum = 0.0D;

   	for (int i = 0; i < this.nrow; i++){
   		for (int j = 0; j < bmat.ncol; j++){
   			sum = 0.0D;
   			for (int k = 0; k < this.ncol; k++){
   				sum += this.matriz[i][k]*bmat.matriz[k][j];
   			}
   			carray[i][j] = sum;
   		}
   	}
   	return cmat;
  }

  // Multiply this matrix by a constant [instance method]
  // This matrix remains unaltered
	//
  public Matriz times(double constant){
   	Matriz cmat = new Matriz(this.nrow, this.ncol);
   	double [][] carray = cmat.getArrayReference();

		for (int i = 0; i < this.nrow; i++){
    	for (int j = 0; j < this.ncol; j++){
    		carray[i][j] = this.matriz[i][j]*constant;
      }
    }
   	return cmat;
  }

  // Multiply two complex matrices {static method]
	//
  public static Matriz times(Matriz amat, Matriz bmat){
   	if (amat.ncol != bmat.nrow)
   		throw new IllegalArgumentException("Nonconformable matrices");

		Matriz cmat = new Matriz(amat.nrow, bmat.ncol);
   	double [][] carray = cmat.getArrayReference();
   	double sum = 0.0D;

   	for (int i = 0; i < amat.nrow; i++){
   		for (int j = 0; j < bmat.ncol; j++){
   			sum = 0.0D;
   			for (int k = 0; k < amat.ncol; k++){
   				sum += (amat.matriz[i][k]*bmat.matriz[k][j]);
   			}
   			carray[i][j] = sum;
   		}
   	}
   	return cmat;
  }

  // Multiply a matrix by a constant [static method]
	//
  public static Matriz times(Matriz amat, double constant){
   	Matriz cmat = new Matriz(amat.nrow, amat.ncol);
   	double [][] carray = cmat.getArrayReference();

    for (int i = 0; i < amat.nrow; i++){
   		for (int j = 0; j < amat.ncol; j++){
   			carray[i][j] = amat.matriz[i][j]*constant;
   		}
   	}
   	return cmat;
  }

  // Multiply this matrix by a matrix [equivalence of *=]
	//
  public void timesEquals(Matriz bmat){
   	if (this.ncol != bmat.nrow)
   		throw new IllegalArgumentException("Nonconformable matrices");

   	double sum = 0.0D;
   	for (int i = 0; i < this.nrow; i++){
   		for (int j = 0; j < bmat.ncol; j++){
   			sum = 0.0D;
   			for (int k = 0; k < this.ncol; k++){
   				sum += (this.matriz[i][k]*bmat.matriz[k][j]);
   			}
   			this.matriz[i][j] = sum;
   		}
   	}
  }

  // Multiply this matrix by a constant [equivalence of *=]
	//
  public void timesEquals(double constant){

   	for (int i = 0; i < this.nrow; i++){
   		for (int j = 0; j < this.ncol; j++){
   			this.matriz[i][j] *= constant;
   		}
   	}
  }

  // INVERSE
  // Inverse of a square matrix [instance method]
	//
  public Matriz inverse(){
   	int n = this.nrow;
   	if (n != this.ncol)
   		throw new IllegalArgumentException("Matriz is not square");
   	double[] col = new double[n];
   	double[] xvec = new double[n];
   	Matriz invmat = new Matriz(n, n);
   	double[][] invarray = invmat.getArrayReference();
   	Matriz ludmat;

   	ludmat = this.luDecomp();
   	for (int j = 0; j < n; j++){
   		for (int i = 0; i < n; i++)
  			col[i] = 0.0D;
   		col[j] = 1.0;
   		xvec = ludmat.luBackSub(col);
   		for (int i = 0; i < n; i++)
   			invarray[i][j] = xvec[i];
   	}
   	return invmat;
  }

  // Inverse of a square matrix [static method]
	//
  public static Matriz inverse(Matriz amat){
   	int n = amat.nrow;
   	if (n != amat.ncol)
   		throw new IllegalArgumentException("Matriz is not square");
   	double[] col = new double[n];
   	double[] xvec = new double[n];
   	Matriz invmat = new Matriz(n, n);
   	double[][] invarray = invmat.getArrayReference();
   	Matriz ludmat;

   	ludmat = amat.luDecomp();
   	for (int j = 0; j < n; j++){
   		for (int i = 0; i < n; i++)
   			col[i] = 0.0D;
   		col[j] = 1.0;
   		xvec = ludmat.luBackSub(col);
   		for (int i = 0; i < n; i++)
   			invarray[i][j] = xvec[i];
   	}
   	return invmat;
  }

  // TRANSPOSE
  // Transpose of a complex matrix [instance method]
	//
  public Matriz transpose(){
  	Matriz tmat = new Matriz(this.ncol, this.nrow);
   	double[][] tarray = tmat.getArrayReference();
   	for (int i = 0; i < this.ncol; i++){
   		for (int j = 0; j < this.nrow; j++){
    			tarray[i][j] = this.matriz[j][i];
   		}
   	}
   	return tmat;
  }

  // Transpose of a matrix [static method]
	//
  public static Matriz transpose(Matriz amat){
   	Matriz tmat = new Matriz(amat.ncol, amat.nrow);
   	double[][] tarray = tmat.getArrayReference();
   	for (int i = 0; i < amat.ncol; i++){
   		for (int j = 0; j < amat.nrow; j++){
   			tarray[i][j] = amat.matriz[j][i];
   		}
  	}
   	return tmat;
  }

  // OPPOSITE
  // Opposite of a matrix [instance method]
	//
  public Matriz opposite(){
   	Matriz opp = Matriz.copy(this);
   	for (int i = 0; i < this.nrow; i++){
   		for (int j = 0; j < this.ncol; j++){
   			opp.matriz[i][j] =- this.matriz[i][j];
   		}
   	}
   	return opp;
	}
  
  // Opposite of a matrix [static method]
	//
  public static Matriz opposite(Matriz amat){
   	Matriz opp = Matriz.copy(amat);
   	for (int i = 0; i < amat.nrow; i++){
   		for (int j = 0; j < amat.ncol; j++){
   			opp.matriz[i][j] =- amat.matriz[i][j];
   		}
   	}
   	return opp;
  }

  // TRACE
  // Trace of a  matrix [instance method]
	//
  public double trace(){
   	double trac = 0.0D;
   	for (int i = 0; i < Math.min(this.ncol,this.ncol); i++){
   		trac += this.matriz[i][i];
  	}
   	return trac;
  }

  // Trace of a matrix [static method]
	//
  public static double trace(Matriz amat){
   	double trac = 0.0D;
   	for (int i = 0; i < Math.min(amat.ncol,amat.ncol); i++){
   		trac += amat.matriz[i][i];
   	}
   	return trac;
  }

  // DETERMINANT
  //  Returns the determinant of a square matrix [instance method]
	//
  public double determinant(){
   	int n = this.nrow;
   	if (n != this.ncol)
   		throw new IllegalArgumentException("Matriz is not square");
   	double det = 0.0D;
   	Matriz ludmat = this.luDecomp();

   	det = ludmat.dswap;
   	for (int j = 0; j < n; j++){
   		det *= ludmat.matriz[j][j];
   	}
   	return det;
  }

  //  Returns the determinant of a square matrix [static method]
	//
  public static double determinant(Matriz amat){
   	int n = amat.nrow;
   	if (n != amat.ncol)
   		throw new IllegalArgumentException("Matriz is not square");
   	double det = 0.0D;
   	Matriz ludmat = amat.luDecomp();

		det = ludmat.dswap;
    for (int j = 0; j < n; j++){
    	det *= (ludmat.matriz[j][j]);
    }
    return det;
  }

  // Returns the log(determinant) of a square matrix [instance method].
  // Useful if determinant() underflows or overflows.
	//
  public double logDeterminant(){
   	int n = this.nrow;
   	if (n != this.ncol)
   		throw new IllegalArgumentException("Matriz is not square");
   	double det = 0.0D;
   	Matriz ludmat = this.luDecomp();

   	det = ludmat.dswap;
   	det=Math.log(det);
   	for (int j = 0; j < n; j++){
   		det += Math.log(ludmat.matriz[j][j]);
   	}
   	return det;
  }

  // Returns the log(determinant) of a square matrix [static method].
  // Useful if determinant() underflows or overflows.
	//
  public static double logDeterminant(Matriz amat){
   	int n = amat.nrow;
   	if (n !=	amat.ncol)
   		throw new IllegalArgumentException("Matriz is not square");
   	double det = 0.0D;
   	Matriz ludmat = amat.luDecomp();

   	det = ludmat.dswap;
   	det = Math.log(det);
   	for (int j = 0; j < n; j++){
   		det += Math.log(ludmat.matriz[j][j]);
   	}
   	return det;
  }

  // FROBENIUS (EUCLIDEAN) NORM of a complex matrix
	//
  public double frobeniusNorm(){
   	double norm = 0.0D;
   	for (int i = 0; i < this.nrow; i++){
   		for (int j = 0; j < this.ncol; j++){
   			norm = hypot(norm, Math.abs(matriz[i][j]));
   		}
   	}
   	return norm;
  }

  // ONE NORM of a matrix
	//
  public double oneNorm(){
  	double norm = 0.0D;
   	double sum = 0.0D;
   	for (int j = 0; j < this.nrow; j++){
   		sum = 0.0D;
   		for (int i = 0; i < this.ncol; i++){
   			sum += Math.abs(this.matriz[i][j]);
   		}
   		norm = Math.max(norm,sum);
   	}
  	return norm;
  }

  // INFINITY NORM of a matrix
	//
  public double infinityNorm(){
   	double norm = 0.0D;
   	double sum = 0.0D;
   	for (int i = 0; i < this.ncol; i++){
   		sum = 0.0D;
   		for (int j = 0; j < this.nrow; j++){
   			sum+=Math.abs(this.matriz[i][j]);
   		}
   		norm = Math.max(norm,sum);
   	}
   	return norm;
  }

  // LU DECOMPOSITION OF COMPLEX MATRIX A
  // For details of LU decomposition
  // See Numerical Recipes, The Art of Scientific Computing
  // by W H Press, S A Teukolsky, W T Vetterling & B P Flannery
  // Cambridge University Press,   http://www.nr.com/
  // Matriz ludmat is the returned LU decompostion
  // int[] index is the vector of row permutations
  // dswap returns +1.0 for even number of row interchanges
  //       returns -1.0 for odd number of row interchanges
	//
  public Matriz luDecomp(){
   	if (this.nrow != this.ncol)
   		throw new IllegalArgumentException("A matrix is not square");
   	int n = this.nrow;
   	int imax = 0;
   	double dum = 0.0D, temp = 0.0D, big = 0.0D;
   	double[] vv = new double[n];
   	double sum = 0.0D;
   	double dumm = 0.0D;

   	this.matrizCheck = true;

		Matriz ludmat = Matriz.copy(this);
   	double[][] ludarray = ludmat.getArrayReference();
	    	
   	ludmat.dswap = 1.0D;
   	for (int i = 0;i < n;i++) {
   		big = 0.0D;
   		for (int j = 0; j < n; j++)
   			if ((temp = Math.abs(ludarray[i][j])) > big)
   				big = temp;
   		if (big == 0.0D){
   			System.out.println("Attempted LU Decomposition of a singular matrix in Matriz.luDecomp()");
   			System.out.println("NaN matrix returned and matrixCheck set to false");
   			this.matrizCheck = false;
   			for (int k = 0;k < n;k++)
   				for (int j = 0;j < n;j++)
   					ludarray[k][j] = Double.NaN;
   			return ludmat;
   		}
   		vv[i] = 1.0/big;
   	}
   	for (int j = 0;j < n;j++) {
   		for (int i = 0;i < j;i++) {
   			sum = ludarray[i][j];
   			for (int k = 0;k < i;k++) 
   				sum -= ludarray[i][k]*ludarray[k][j];
   			ludarray[i][j] = sum;
   		}
   		big = 0.0D;
   		for (int i = j;i < n;i++) {
   			sum = ludarray[i][j];
   			for (int k = 0;k < j;k++)
   				sum -= ludarray[i][k]*ludarray[k][j];
   			ludarray[i][j] = sum;
   			if ((dum = vv[i]*Math.abs(sum)) >= big) {
   				big = dum;
   				imax = i;
   			}
   		}
   		if (j != imax) {
   			for (int k = 0;k < n;k++) {
   				dumm = ludarray[imax][k];
   				ludarray[imax][k] = ludarray[j][k];
   				ludarray[j][k] = dumm;
   			}
   			ludmat.dswap = -ludmat.dswap;
   			vv[imax] = vv[j];
   		}
   		ludmat.index[j] = imax;

   		if (ludarray[j][j] == 0.0D){
   			ludarray[j][j] = TINY;
   		}
   		if (j != n-1) {
   			dumm = 1.0/ludarray[j][j];
   			for (int i = j+1;i < n;i++){
   				ludarray[i][j] *= dumm;
   			}
   		}
   	}
   	return ludmat;
  }

  // Solves the set of n linear equations A.X=B using not A but its LU decomposition
  // bvec is the vector B (input)
  // xvec is the vector X (output)
  // index is the permutation vector produced by luDecomp()
	//
  public double[] luBackSub(double[] bvec){
   	int ii = 0,ip = 0;
   	int n = bvec.length;
   	if (n != this.ncol)
   		throw new IllegalArgumentException("vector length is not equal to matrix dimension");
   	if (this.ncol != this.nrow)
   		throw new IllegalArgumentException("matrix is not square");
   	double sum =  0.0D;
   	double[] xvec = new double[n];
   	for (int i = 0; i < n; i++){
   		xvec[i] = bvec[i];
   	}
  	for (int i = 0;i < n;i++) {
   		ip = this.index[i];
   		sum = xvec[ip];
   		xvec[ip] = xvec[i];
   		if (ii == 0){
   			for (int j = ii; j <= i-1;j++){
   				sum -= this.matriz[i][j]*xvec[j];
   			}
   		} else{
   			if (sum == 0.0) ii=i;
   		}
   		xvec[i] = sum;
   	}
   	for (int i = n-1;i >= 0;i--) {
   		sum = xvec[i];
   		for (int j = i+1;j < n;j++){
   			sum -= this.matriz[i][j]*xvec[j];
   		}
   		xvec[i] =  sum/matriz[i][i];
   	}
   	return xvec;
  }
    
  // Solves the set of n linear equations A.X=B
  // bvec is the vector B (input)
  // xvec is the vector X (output)
	//
  public double[] solveLinearSet(double[] bvec){
   	Matriz ludmat =	this.luDecomp();

   	return ludmat.luBackSub(bvec);
  }

  // Real methods not in java.lang.maths required in this Class
  // See Fmath.class for public versions of these methods
	//
  private static double hypot(double aa, double bb){
   	double cc = 0.0D, ratio = 0.0D;
   	double amod = Math.abs(aa);
   	double bmod = Math.abs(bb);
    	
   	if (amod == 0.0D){
   		cc = bmod;
   	} else {
   		if (bmod == 0.0D){
   			cc = amod;
   		} else {
   			if (amod <= bmod){
   				ratio = amod/bmod;
   				cc = bmod*Math.sqrt(1.0D+ratio*ratio);
   			} else {
   				ratio = bmod/amod;
   				cc = amod*Math.sqrt(1.0D+ratio*ratio);
   			}
   		}
   	}
   	return cc;
  }

	/*
			MODIFICAES INTRODUZIDAS POR:
			JOS IGUELMAR MIRANDA
	*/

  // Calcula autovalores e autovetores de uma matriz quadrada
	// Retorna um objeto Matriz com os 'eigen':
	// (1) eigenvalues na primeira linha e 
	// (2) eigenvectors da segunda linha em diante.
	// Fonte: Press et al.
	//
  public Matriz autoValores() {

		int nlin, ncol;

	 	nlin = this.nrow;
		ncol = this.ncol;

		// Certifica que a matriz  quadrada
		//
 		if (nlin != ncol) 
   		throw new IllegalArgumentException("A matriz no  quadrada.");

   	Matriz m_auto = new Matriz(nlin + 1, ncol);
   	eigen = m_auto.getArrayReference();
		m_auto.nrow = nlin + 1;
		m_auto.ncol = ncol;

		double[] d = new double[nlin],						// d = eigenValues
		  			 b = new double[nlin],
   					 z = new double[nlin];
   	int nrot, j, iq, ip, i, numBand;
   	double tresh, theta, tau, t, sm, s, h, g, c;
   	double[][] 	v = new double[nlin][ncol], 		// v = eigenVectors
		    				a_array = new double[nlin][ncol];
    	
   	a_array = this.getArrayCopy();
 	 	numBand = ncol;
   	   	
 	 	// Cria matriz indentidade
		//
   	for (ip = 1; ip <= numBand; ip++) {
   		for (iq = 1; iq <= numBand; iq++)
   			v[ip-1][iq-1] = 0.0d;
   		v[ip-1][ip-1] = 1.0d;
   	}
    	
   	// Inicializa b e d para a diagonal de 'a' (a_array)
		//
   	for (ip = 1; ip <= numBand; ip++) {
   		b[ip-1] = a_array[ip-1][ip-1];
   		d[ip-1] = a_array[ip-1][ip-1];
   		z[ip-1] = 0.0d;
   	}
   	nrot = 0;
   	for (i = 1; i < 50; i++) {
   		sm = 0.0d;
   		for (ip = 1; ip <= numBand - 1; ip++) {
   			for (iq = ip + 1; iq <= numBand; iq++) 
   	            sm += Math.abs(a_array[ip-1][iq-1]);
   		}
   		if (sm == 0.0) {

   			// Finalizou operao
				//
   			for (int k = 0; k < numBand; k++)
   				eigen[0][k] = d[k];
   			for (int m = 0; m < numBand; m++) {
     			for (int n = 0; n < numBand; n++) {
     				eigen[m+1][n] = v[m][n];
     			}
   			}
   			return m_auto;
   		}
   		if (i < 4) {
   			tresh = 0.2*sm/(double)(numBand*numBand);
   		} else
         tresh = 0.0d;
   		for (ip = 1; ip <= numBand - 1; ip++) {
   			for (iq = ip + 1; iq <= numBand; iq++) {
   				g = 100.0*Math.abs(a_array[ip-1][iq-1]);
          if ((i > 4) && ((Math.abs(d[ip-1]) + g) == Math.abs(d[ip-1])) &&
             ((Math.abs(d[iq-1]) + g) == Math.abs(d[iq-1]))) {
             a_array[ip-1][iq-1] = 0.0d;
          } else if (Math.abs(a_array[ip-1][iq-1]) > tresh) {
             h = d[iq-1] - d[ip-1];
             if ((Math.abs(h) + g) == Math.abs(h)) {
                t = a_array[ip-1][iq-1]/h;
             } else {
                theta = 0.5*h/a_array[ip-1][iq-1];
                t = 1.0d/(Math.abs(theta) + Math.sqrt(1.0d + theta*theta));
                if (theta < 0.0) 
                   t = -t;
             }
             c = 1.0/Math.sqrt(1.0 + t*t);
             s = t*c;
             tau = s/(1.0d + c);
             h = t*a_array[ip-1][iq-1];
             z[ip-1] = z[ip-1] - h;
             z[iq-1] = z[iq-1] + h;
             d[ip-1] = d[ip-1] - h;
             d[iq-1] = d[iq-1] + h;
             a_array[ip-1][iq-1] = 0.0d;
             for (j = 1; j <= ip - 1; j++) {

           	   // Rotate(a,j,ip,j,iq)
							 //
           	   g = a_array[j-1][ip-1];
           	   h = a_array[j-1][iq-1];
           	   a_array[j-1][ip-1] = g - s*(h + g*tau);
               a_array[j-1][iq-1] = h + s*(g - h*tau);
             }
             for (j = ip + 1; j <= iq - 1; j++) {

		       	   // Rotate(a,ip,j,j,iq);
							 //
           	   g = a_array[ip-1][j-1];
           	   h = a_array[j-1][iq-1];
           	   a_array[ip-1][j-1] = g - s*(h + g*tau);
           	   a_array[j-1][iq-1] = h + s*(g - h*tau);
             }
             for (j = iq + 1; j <= numBand; j++) {

           	   // Rotate(a,ip,j,iq,j);
							 //
           	   g = a_array[ip-1][j-1];
           	   h = a_array[iq-1][j-1];
           	   a_array[ip-1][j-1] = g - s*(h + g*tau);
           	   a_array[iq-1][j-1] = h + s*(g-h*tau);
             }
             for (j = 1; j <= numBand; j++) {

               // Rotate(v,j,ip,j,iq);
							 //
               g = v[j-1][ip-1];
               h = v[j-1][iq-1];
               v[j-1][ip-1] = g - s*(h+g*tau);
               v[j-1][iq-1] = h + s*(g-h*tau);
             }
          }	// else if
    	    nrot++;
    		}
    	}		// for (ip = 1; ip <= numBand - 1; ip++)
    	for (ip = 1; ip <= numBand; ip++) {
    		b[ip-1] = b[ip-1] + z[ip-1];
    		d[ip-1] = b[ip-1];
    		z[ip-1] = 0.0d;
    	}
    }		// for (i = 1; i < 50; i++)
    System.out.println("eigenValue: Muitas iteraes na rotina Jacobi.");
    return null;
  }

	// Retorna apenas os autovalores da matriz
	//
	public Matriz retornaValores() {
		int nlin, ncol;

		// Lembrar que a Matriz de autovalores tem lin+1 linhas!
		//
	 	nlin = this.nrow - 1;
	 	ncol = this.ncol;

   	double[][] a_autoValores = new double[nlin][1], 
							 a_eigen = new double[nlin+1][ncol];

		// Salva a cpia do eigen antes de criar novo objeto Matriz e
		// consequente criao de um novo array eigen vazio!
		//
   	a_eigen = this.getArrayCopy();
   	Matriz m_autoValores = new Matriz(a_autoValores);

		// Cria Matriz com autovalores.
		// Os autovalores esto armazenados na primeira linha do array
		// de autovalores.
		//
		for (int i = 0; i < nlin; i++) {
		    a_autoValores[i][0] = a_eigen[0][i];
		}
		return m_autoValores;
	}

	// Retorna apenas os autovetores da matriz
	//
	public Matriz retornaVetores() {
		int nlin, ncol;

		// Lembrar que a Matriz de autovalores tem lin+1 linhas!
		//
	 	nlin = this.nrow - 1;
	 	ncol = this.ncol;
   	double[][] a_autoVetores = new double[nlin][ncol], 
							 a_eigen = new double[nlin+1][ncol];

		// Salva a cpia do eigen antes de criar novo objeto Matriz e
		// consequente criao de um novo array eigen vazio!
		//
   	a_eigen = this.getArrayCopy();
   	Matriz m_autoVetores = new Matriz(a_autoVetores);

		// Cria Matriz com autovetores.
		// Os autovetores esto armazenados da segunda linha em 
		// diante do array de autovalores.
		//
		for (int i = 0; i < nlin; i++) 
			for (int j = 0; j < ncol; j++)
				a_autoVetores[i][j] = a_eigen[i+1][j];

		return m_autoVetores;
	}

  // Imprime a matriz
	//
  public void imprimeMatriz() {
    	
		String padrao = "#.###";
		DecimalFormat formato = null;
		formato = new DecimalFormat(padrao);
		String elemento;

		if (this.matriz == null){
   		System.out.println("Matriz vazia.");
   	} else {
   		int nr = this.getNrow();
   		int nc = this.getNcol();
   		for (int i = 0; i < nr; i++){
   			for (int j = 0; j < nc; j++){
   				elemento = formato.format(this.matriz[i][j]);
   				System.out.print(elemento+ "\t");
   			}
				System.out.println();
   		}
   	}
  }

  // Imprime transposta da matriz
	//
  public void imprimeTransposta() {
    	
		String padrao = "#.###";
		DecimalFormat formato = null;
		formato = new DecimalFormat(padrao);
		String elemento;

		if (this.matriz == null){
   		System.out.println("Matriz vazia.");
   	} else {
				Matriz tmat = new Matriz(this.ncol, this.nrow);
				double[][] tarray = tmat.getArrayReference();
				for (int i = 0; i < this.ncol; i++){
					for (int j = 0; j < this.nrow; j++){
							tarray[i][j] = this.matriz[j][i];
					}
				}
				for (int i = 0; i < this.ncol; i++){
					for (int j = 0; j < this.nrow; j++){
						elemento = formato.format(tarray[i][j]);
						System.out.print(elemento+ "\t");
					}
					System.out.println();
				}
   	}
  }

  // Imprime a matriz com 'decimais' casas decimais
	//
  public void imprimeMatriz(int decimais) {

		String padrao = "#.";
		for (int i = 0; i < decimais; i++)
			padrao += "#";
		DecimalFormat formato = null;
		formato = new DecimalFormat(padrao);
		String elemento;

		if (this.matriz == null){
   		System.out.println("Matriz vazia.");
   	} else {
   		int nr = this.getNrow();
   		int nc = this.getNcol();
   		for (int i = 0; i < nr; i++){
   			for (int j = 0; j < nc; j++){
   				elemento = formato.format(this.matriz[i][j]);
   				System.out.print(elemento+ "\t");
   			}
				System.out.println();
   		}
   	}
  }

  // Imprime a matriz inferior
	//
  public void imprimeMatrizInferior() {
    	
		String padrao = "#.###";
		DecimalFormat formato = null;
		formato = new DecimalFormat(padrao);
		String elemento;

   	if (this.nrow != this.ncol){
   		throw new IllegalArgumentException("Array dimensions do not agree");
   	}
		if (this == null){
   		System.out.println("Matriz vazia.");
   	} else {
   		int nr = this.getNrow();
   		int nc = this.getNcol();
   		for (int i = 0; i < nr; i++){
   			for (int j = 0; j < nc; j++){
   				if (i >= j) {
	   				elemento = formato.format(this.matriz[i][j]);
   					System.out.print(elemento+ "\t");
   				}
   			}
				System.out.println();
   		}
   	}
  }

  // Ordena autovalores
	//
  public int[] ordenaAutoValores() {
		  	    	
   	// Reverte de Matriz para array
		//
    double[][] arrayValor = this.getArrayReference();
    double autoValores[];
    int j = 0, ordem[], nlin = this.getNrow();

		// Matriz a ser retornada com ordem dos elementos
		//
    ordem = new int[nlin];
    	
    // Copia valores da Matriz de autovalores para array local
		//
    autoValores = new double[nlin];
    for (int i = 0; i < nlin; i++)
    	autoValores[i] = this.getElement(i, 0);
    
    // Ordena array local em ordem ascendente
		//
    Arrays.sort(autoValores);
    	  	
    // Troca para ordem descendente
		//
    double temp;
    int indice = nlin - 1;
    for (int i = 0; i < indice/2; i++) {
    	temp = autoValores[i];
    	autoValores[i] = autoValores[indice - i];
    	autoValores[indice - i] = temp;
    }

    // Acha indices na Matriz de autovalores inicial 
		//
    for (int i = 0; i < nlin; i++) {
    	while (autoValores[i] != this.getElement(j, 0))
    		++j;
    	ordem[i] = j;
    	j = 0;
    }
    	
    // Copia array local para Matriz de autovalores
		//
    for (int i = 0; i < nlin; i++) 
    	arrayValor[i][0] = autoValores[i];
    	    	
    return ordem;
  }

  // Ordena autovetores
	//
  public void ordenaAutoVetores(int[] ordem) {
    	    	
  	// Reverte de Matriz para array
		//
    double[][] arrayVetor = this.getArrayReference();
   	
    // Ordena Matriz de autovetores de acordo com ordem dos autovalores, 
		// trocando as colunas
		//
   	int nlin = this.getNrow();
    	
   	double[][] tempArray = new double[nlin][nlin];
   	for (int col = 0; col < nlin; col++) {
   		for (int lin = 0; lin < nlin; lin++) {
   			tempArray[lin][col] = arrayVetor[lin][ordem[col]]; 
   		}
   	}
    	
   	// Copia array para Matriz
		//
   	for (int col = 0; col < nlin; col++) {
   		for (int lin = 0; lin < nlin; lin++) {
   			arrayVetor[lin][col] = tempArray[lin][col]; 
   		}
   	}  	
  }

  // Calcula porcentagem da varincia total explicada por cada 
	// componente principal (autovalor)
	//
  public double[] porcentagemVariancia() {
    	    	
   	double[] pVariancia;

   	// Reverte de Matriz para array
		//
   	double[][] arrayValor = this.getArrayReference();
   	double soma = 0.0d;
   	
   	// Ordena Matriz de autovetores de acordo com ordem dos autovalores, 
		// trocando as colunas
		//
   	int nlin = this.getNrow();
    	
   	pVariancia = new double[nlin];
   	for (int lin = 0; lin < nlin; lin++) {
   		soma += arrayValor[lin][0];
   	}
   	for (int lin = 0; lin < nlin; lin++) {
   		pVariancia[lin] = (arrayValor[lin][0]*100)/soma;
   	}
   	return pVariancia;
  }

  // Calcula grau de correlao entre componente principal e banda da imagem
	//
  public void correlacaoComponentePrincipal(double[] variancia,	Matriz autoVal, 
			Matriz autoVet) {
		  	    	   	
   	// Reverte de Matriz para array
		//
   	double[][] arrayCorrel = this.getArrayReference();
   	double[][] a_autoVal = autoVal.getArrayCopy();
   	double[][] a_autoVet = autoVet.getArrayCopy();
   	
   	int nlin = this.getNrow();
   	
   	for (int k = 0; k < nlin; k++) {
   		for (int p = 0; p < nlin; p++) {
   			arrayCorrel[k][p] = (a_autoVet[k][p]*Math.sqrt(a_autoVal[p][0]))/Math.sqrt(variancia[k]);
   		}
   	}     	
  }

	// Normaliza (standardize) os valores de uma matriz
	//
  public void normaliza() {

   	int nlin = this.getNrow(), ncol = this.getNcol();
   	double t, u, v, q, s;
    	
   	for (int k = 0; k < ncol; k++) {
  		t = 0.0;
   		u = 0.0;
   		for (int i = 0; i < nlin; i++) {
   			v = this.getElement(i, k);
   			t += v;
   			u += v*v;
   		}
   		q = t/(double)nlin;
   		s = Math.sqrt(((double)nlin - 1.0)/(u - t*q));
   		for (int i = 0; i < nlin; i++) 
   			this.setElement(i, k, s*(this.getElement(i, k) - q));
   	}
  }

	// Calcula distncia Euclidiana.
	// As variveis devem estar nas colunas e os casos nas linhas
	// O resultado  retornado numa Matriz quadrada, com dimenso casos x casos
	//
	public Matriz distEuclidiana() {
		int lin, col;
		double diff = 0.0, x;

	 	lin = this.nrow;
		col = this.ncol;
   	Matriz m_dist = new Matriz(lin, lin);
   	double[][] a_dist = m_dist.getArrayReference();
		m_dist.nrow = lin;
		m_dist.ncol = lin;

		// Cria Matriz de distncia Euclidiana
		//
		for (int i = 0; i < lin; i++) {
			for (int k = 0; k < lin; k++) {
				for (int j = 0; j < col; j++) {
					x = this.matriz[i][j] - this.matriz[k][j];
					diff += x*x;
				}
				a_dist[i][k] = Math.sqrt(diff);
				diff = 0.0;
			}
		}
		return m_dist;
	}

	// Calcula distncia Euclidiana ponderada
	// As variveis devem estar nas colunas e os casos nas linhas
	// O resultado  retornado numa Matriz, com dimenso casos x casos
	// Fonte: Hartigan (1975), pg. 60.
	//
	public Matriz distEuclidianaPonderada() {
		int lin, col, N;
		double diff = 0.0, x = 0.0, sum = 0.0, sum2 = 0.0;
		double[] variancia;

	 	lin = this.nrow;
		col = this.ncol;
		variancia = new double[ncol];
   	Matriz m_dist = new Matriz(lin, lin);
   	double[][] a_dist = m_dist.getArrayReference();
		m_dist.nrow = lin;
		m_dist.ncol = lin;

		// Calcula valores de varincia das variveis
		//
		for (int j = 0; j < col; j++) {
			for (int i = 0; i < lin; i++) {
				x = this.matriz[i][j];
				sum2 += x*x;
				sum += x;
			}
			variancia[j] = (double)1/(lin-1)*(sum2 - ((sum*sum)/lin));
			sum2 = 0.0;
			sum = 0.0;
		}

		// Cria Matriz de distncia Euclidiana ponderada
		// O peso  o inverso da varincia (Hartigan, 1975:60)
		//
		for (int i = 0; i < lin; i++) {
			for (int k = 0; k < lin; k++) {
				for (int j = 0; j < col; j++) {
					x = this.matriz[i][j] - this.matriz[k][j];
					diff += (1/variancia[j])*x*x;
				}
				a_dist[i][k] = Math.sqrt(diff);
				diff = 0.0;
			}
		}
		return m_dist;
	}
}

